"""
"""

goto_host_func_ = "goto_host_func"
goto_service_func_ = "goto_service_func"
goto_repo_func_ = "goto_repo_func"

desc_host_func_ = "desc_host_func"
desc_service_func_ = "desc_service_func"

list_host_func_ = "list_host_func"
list_service_func_ = "list_service_func"

desc_commit_func_ = "desc_commit_func"
